"use strict";

(function() {

  /// ユーティリティ
  let U = {
    init: function() {
    }
    /// 新しいIDを生成する
  , newId: function() {
      return Math.floor(Math.random() * 100000);
    }
  };

  /// モデル（コアとなるデータとロジック）
  let M = (function() {
    function init() {
    }
    /// タスク
    class Task {
      constructor(id, title, done) {
        this.id    = id;
        this.title = title;
        this.done  = !! done;
      }
    }
    /// タスクリスト
    class TaskList {
      constructor() {
        this._list = [];
      }
      /// タスクを追加
      add(title, done) {
        let task = new Task(U.newId(), title, done);
        this._list.push(task);
        return task;
      }
      /// タスクをIDで検索
      get(id) {
        return _.find(this._list, (x) => (x.id === id));
      }
      /// タスクの完了済フラグを切り替える
      toggle(id) {
        let task = this.get(id);
        if (task) {
          task.done = ! task.done;
        }
        return task;
      }
      /// すべてのタスクの完了済フラグを切り替える
      toggleAll(done, callback) {
        for (let task of this._list) {
          if (task.done !== done) {
            task.done = done;
            if (callback) { callback(task); }
          }
        }
      }
      /// タスクを削除する
      delete(id) {
        let idx = _.findIndex(this._list, (x) => (x.id === id));
        if (idx < 0) { return null; }
        let task = this._list[idx];
        this._list.splice(idx, 1);
        return task;
      }
      /// 完了済のタスクをすべて削除する
      clean() {
        let pair = _.partition(this._list, (x)=>(x.done === true));
        this._list = pair[1];  // 未完了のタスクだけを残す
        return pair[0];        // 完了済のタスクだけを返す
      }
      /// タスクのタイトルを更新する
      update(id, title) {
        let task = this.get(id);
        if (! task) { throw "task not found (id="+id+")."; }
        task.title = title;
        return task;
      }
    }
    //
    return {init: init, tasklist: new TaskList()};
  })();

  /// コントローラ（イベントハンドラ）
  let C = {
    init: function() {
      C.registerHandlers();           // イベントハンドラを登録
      C._editing = null;              // 編集中のタスクID
    }
    /// イベントハンドラを登録する
  , registerHandlers: function() {
      $('#add-task').on('submit', C.doAddTask);
      $('#todo-list')
        .on('click'   , 'input[type=checkbox]', C.doToggleTask)
        .on('click'   , 'button.destroy'      , C.doDeleteTask)
        .on('dblclick', 'label'               , C.doEditTask)
        .on('submit'  , 'form'                , C.doUpdateTask)
      ;
      $('#toggle-all').on('click', C.doToggleAllTasks);
      $('#clear-completed').on('click', C.doCleanTasks);
    }
    /// タスクを追加する
  , doAddTask: function() {
      /// thisは<form>なので、その子孫要素からテキストフィールドを探す
      let $input = $(this).find('input[type=text][name=title]');
      /// テキストフィールドが空でなければ、タスクを追加
      let title = $input.val().trim();
      if (title) {
        let task = M.tasklist.add(title);
        V.addTask(task);
      }
      $input.val("");  /// テキストフィールドを空にする
      return false;    /// フォームデータを送信しない
    }
    /// 操作対象のタスクIDを返す
  , _getTaskId: function(elem) {
      let id = $(elem).parents('.task').data('id');
      if (! id) { throw "failed to get task id."; }
      return +id;   // 文字列を数値に変換
    }
    /// タスクが未完了なら完了済へ、完了済なら未完了へ切り替える
  , doToggleTask: function() {
      let id = C._getTaskId(this);
      let task = M.tasklist.toggle(id);
      V.toggleTask(task);
    }
    /// すべてのタスクを完了済または未完了へ変更する
  , doToggleAllTasks: function() {
      let done = this.checked; // true: close, false: open
      //if (done) {
      //  $('.task input[type=checkbox]:not(:checked)').trigger('click');
      //} else {
      //  $('.task input[type=checkbox]:checked').trigger('click');
      //}
      M.tasklist.toggleAll(done, function(task) {
        V.toggleTask(task);
      });
    }
    /// タスクを削除する
  , doDeleteTask: function() {
      let id = C._getTaskId(this);
      let task = M.tasklist.delete(id);
      V.deleteTask(task);
    }
    /// 完了済のタスクをすべて削除する
  , doCleanTasks: function() {
      //$('.task.completed').each(function() {
      //  let id = +$(this).data('id');
      //  M.tasklist.delete(id);
      //}).fadeOut();
      let deletedTasks = M.tasklist.clean();
      for (let task of deletedTasks) {
        V.deleteTask(task);
      }
    }
    /// タスクの編集を開始する
  , doEditTask: function() {
      if (C._editing && M.tasklist.get(C._editing)) {
        alert("Please update editing task.");
        return;
      }
      let id = C._getTaskId(this);
      C._editing = id;
      V.editTask(M.tasklist.get(id));
    }
    /// タスクの編集を終了する
  , doUpdateTask: function() {
      let title = $(this).find('input.edit').val().trim();
      if (! title) {
        alert("Task title requried.");
      } else {
        let id = C._getTaskId(this);
        let task = M.tasklist.update(id, title);
        V.updateTask(task);
        C._editing = null;
      }
      return false;
    }
  };

  /// ビュー（表示に関する機能）
  let V = {
    init: function() {
    }
    /// タスクのDOM要素（<li>タグ）を作る
  , renderTask: function(task) {
      let id  = 'task-' + task.id;
      let cls = task.done ? 'task completed' : 'task';
      let arr =
        ['li', {'id': id, 'class': cls, 'data-id': task.id}
        , ['div', {'class': 'view'},
          , ['input', {'class': 'toggle', type: 'checkbox'}]
          , ['label', task.title]
          , ['button', {'class': 'destroy'}]
          ]
        , ['form'
          , ['input.edit', {type: 'text', value: task.title}]
          ]
        ];
      return a2dom(arr);
    }
    /// タスクリストに新しいタスクを追加する（<li>タグ）
  , addTask: function(task) {
      $('#todo-list').prepend(V.renderTask(task));
    }
    /// タスクの表示を完了済に変更する
  , checkTask: function(task) {
      $('#task-'+task.id).addClass('completed')
        .find('input[type=checkbox]').prop('checked', true);
    }
    /// タスクの表示を未完了に戻す
  , uncheckTask: function(task) {
      $('#task-'+task.id).removeClass('completed')
        .find('input[type=checkbox]').prop('checked', false);
    }
    /// タスクの表示を、完了済なら未完了へ、未完了なら完了済へ変更
  , toggleTask: function(task) {
      task.done ? V.checkTask(task) : V.uncheckTask(task);
    }
    /// タスクリストからタスクを取り除く
  , deleteTask: function(task) {
      //$('#task-'+task.id).remove();  // 取り除くだけ
      $('#task-'+task.id).fadeOut();   // アニメーションつき
    }
    /// タスクを編集状態にする
  , editTask: function(task) {
      $('#task-'+task.id).addClass('editing');
    }
    /// タスクの表示を更新する
  , updateTask: function(task) {
      let elem = V.renderTask(task);
      $('#task-'+task.id).replaceWith(elem);
    }
  };

  /// 初期設定用
  function init() {
    U.init();
    M.init();
    V.init();
    C.init();
  }

  /// 初期設定を行う
  if (typeof $ !== 'undefined') { $(init); }

})();

if (typeof module !== 'undefined') { module.exports = TodoApp; }
